/*
 * Copyright (C) 2001/2003 The Honeynet Project.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by The Honeynet Project.
 * 4. The name "The Honeynet Project" may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#pragma ident "@(#)$Id: net.c,v 1.7 2003/11/30 18:37:58 job Exp $"

#include "sebek.h"

struct qinit *old_rqinit = NULL;
struct qinit *old_wqinit = NULL;

struct qinit new_rqinit;
struct qinit new_wqinit;

queue_t *qinp;
queue_t *qoutp;

struct ether_addr ether_src;
struct ether_addr ether_dst;
struct vnode *sbkvp;

int32_t packet_cntr = 0;

struct in_addr ip_local;
struct in_addr ip_dst;

int iprput(q, mp)
queue_t *q;
mblk_t *mp;
{
    int (*nextput)(queue_t *, mblk_t *);
    ipha_t *ipha;
    int len;
    ipaddr_t dst;
    unsigned char *rp;
    struct udphdr *uh;
    ipha_t *iph;
    uint32_t up;

    if (MTYPE(mp) != M_DATA) 
        sebek_log(SEBEK_CE_DEBUG1, "incoming packet %d, base %lx, rptr %lx\n",MTYPE(mp), mp->b_datap->db_base, mp->b_rptr);

    iph = (ipha_t *)(mp->b_rptr);
    uh = (struct udphdr*)(mp->b_rptr + sizeof(ipha_t));
    rp = mp->b_rptr + sizeof(ipha_t) + sizeof(struct udphdr);
    if (((rp+4) <= mp->b_wptr) && (iph->ipha_protocol == htonl(IPPROTO_UDP)) && (uh->uh_dport == htons(DESTINATION_PORT))) {
        bcopy(rp, &up, 4);
        if (ntohl(up) == MAGIC_VAL) {
            sebek_log(SEBEK_CE_DEBUG2, "hiding packet in ip stream!\n");
            mp->b_prev = NULL;
            freemsg(mp);
            return(0);
        }
    }

    nextput = old_rqinit->qi_putp;

    if (nextput) {
        return (*nextput)(q, mp);
    }
    return(0);
}

int ipwput(q, mp)
queue_t *q;
mblk_t *mp;
{
    int (*nextput)(queue_t *, mblk_t *);

    sebek_log(SEBEK_CE_DEBUG2, "outgoing packet\n");

    nextput = old_wqinit->qi_putp;

    if (nextput) {
        return (*nextput)(q, mp);
    }
    return(0);
}

int new_strrput(queue_t * q, mblk_t * mp)
{
    int len;
    int idnum;

    /*
     * Try to limit the overhead as much as possible. Still pretty ugly
     * to catch every strrput only for raw network traffic.
     */

    if (!(mp && mp->b_datap)) {
        return (strrput(q, mp));
    }

    if (!(WR(q)->q_next && WR(q)->q_next->q_qinfo && WR(q)->q_next->q_qinfo->qi_minfo)) {
        return (strrput(q, mp));
    }

    idnum = WR(q)->q_next->q_qinfo->qi_minfo->mi_idnum;


    /* We want to intercept data not from these */
    if (idnum == 44 || idnum == 43 || idnum == 57005) {
          return (strrput(q, mp));
    }

    sebek_log(SEBEK_CE_DEBUG1, "packet from %d type %x\n", idnum, MTYPE(mp));
 
/* Debug stuff that is only valid for sparc

    if (idnum == HMEIDNUM) {
        switch (mp->b_datap->db_type) {
            case M_DATA:
                        break;
            case M_PROTO:
            case M_PCPROTO:
                sebek_log(SEBEK_CE_DEBUG1, "primitive %d / %x\n",
                    ((dl_unitdata_ind_t *)(mp->b_rptr))->dl_primitive,
                    ((dl_unitdata_ind_t *)(mp->b_rptr))->dl_primitive);
       }
    }
*/
    return (strrput(q, mp));
}

int send_packet(char *buf, unsigned int len, int fd, int fromuser, uint16_t type)
{
    char *dst;
    struct ether_header *ehp;
    ipha_t iph, *iphp;
    struct udphdr uh, *uhp;
    struct sbk_h sbkh, *sbkhp;
    mblk_t *mp;
    int32_t cntr;
    int err;
    struct timeval32 tv;
    proc_t *p;


    sebek_log(SEBEK_CE_DEBUG3, "send_packet: %d len enter", len);
    if (!stream_open || len > 1500) return(-1);

    /*
     * Send the data including ethernet header with an M_DATA packet.
     */
    mp = makemblk(sizeof(struct ether_header)+sizeof(ipha_t)+sizeof(struct udphdr)+sizeof(struct sbk_h)+len, M_DATA);
    ehp = (struct ether_header *)(mp->b_rptr);
    bcopy(&ether_src, &ehp->ether_shost, ETHERADDRL);
    bcopy(&ether_dst, &ehp->ether_dhost, ETHERADDRL);
    ehp->ether_type =(u_short) htons(0x800);
  
    /*
     * IP header fields copied from original sebek2
     */ 
    iph.ipha_version_and_hdr_length = IP_SIMPLE_HDR_VERSION;
    iph.ipha_type_of_service = 13;
    iph.ipha_length = htons(len + 8 + 20);
    iph.ipha_ident = 0;
    iph.ipha_fragment_offset_and_flags = 0;
    iph.ipha_ttl = 32;
    iph.ipha_protocol = IPPROTO_UDP;
    iph.ipha_hdr_checksum = 0;
    iph.ipha_src = ip_local.s_addr;
    iph.ipha_dst = ip_dst.s_addr;
    iph.ipha_hdr_checksum = ip_csum_hdr(&iph);

    iphp = (ipha_t *)(ehp + 1); 
    bcopy(&iph, iphp, sizeof(ipha_t));

    uh.uh_sport = htons(SOURCE_PORT);
    uh.uh_dport = htons(DESTINATION_PORT);
    uh.uh_ulen = htons(sizeof(struct udphdr) + sizeof(struct sbk_h) + len);
    uh.uh_sum = 0;

    uhp = (struct udphdr *)(iphp + 1);
    bcopy(&uh, uhp, sizeof(struct udphdr));

    uniqtime32(&tv);
    p = ttoproc(curthread);

    sbkh.magic = htonl(MAGIC_VAL);
    sbkh.ver = htons(SBK_VER);
    sbkh.type = htons(type);
    mutex_enter(&sebek_lock);
    sbkh.counter = htonl(packet_cntr++);
    mutex_exit(&sebek_lock);
    sbkh.time_sec = htonl(tv.tv_sec);
    sbkh.time_usec = htonl(tv.tv_usec);
    sbkh.pid = htonl(p->p_pid);
    sbkh.uid = htonl(p->p_cred->cr_ruid);
    sbkh.fd = htonl(fd);
    strncpy(sbkh.com, p->p_user.u_comm, 12);
    sbkh.length = htonl(len);

    sbkhp = (struct sbk_h *)(uhp + 1);
    bcopy(&sbkh, sbkhp, sizeof(struct sbk_h));

    dst = (char *)(sbkhp + 1);

    sebek_log(SEBEK_CE_DEBUG3, "send_packet ptrs %lx, %lx\n",ehp,dst);

    if (fromuser) {
        copyin(buf, dst, len);
        sebek_log(SEBEK_CE_DEBUG2, "sending cnt %d\n", packet_cntr);
    } else {
        bcopy(buf, dst, len);
        sebek_log(SEBEK_CE_DEBUG2, "sending cnt %d, %s\n", packet_cntr, buf);
    }

    if ((err = kstrputmsg(sbkvp, mp, NULL, 0, 0, MSG_HIPRI, 0)) != 0) {
        sebek_log(SEBEK_CE_DEBUG1, "send failed:\n",err);
        return(-1);
    }

    return(0);
}

mblk_t *makemblk(unsigned int size, int type)
{
    mblk_t *mp;
    char *cp;

    mp = allocb(size, BPRI_HI);
    mp->b_datap->db_type = type;
    cp = (char *)mp->b_rptr;
    mp->b_wptr = (u_char *)&cp[size];
    bzero(cp, size);

    return(mp);
}

int send_and_ack(struct vnode *vp, mblk_t **mpp, rval_t *retvalp)
{
    unsigned char pri;
    int pflag;
    int timeout;
    int err;

    if ((err = kstrputmsg(vp, *mpp, NULL, 0, 0, MSG_BAND, 0)) != 0) {
        return(err);
    }

    *mpp = NULL;
    timeout = 0;
    pflag = MSG_ANY;
    pri = 0;

    err = kstrgetmsg(vp, mpp, NULL, &pri, &pflag, timeout, retvalp);

    if (err) {
        sebek_log(SEBEK_CE_DEBUG1, "ack error %d\n",err);
        return(err);
    }

    return(0);
}

/*
 * Taken from libpcap.
 * Convert 's' which has the form "xx:xx:xx:xx:xx:xx" into a new
 * ethernet address.  Assumes 's' is well formed.
 */

/* Hex digit to integer. */
static int xdtoi(int c)
{
        if (isdigit(c))
                return c - '0';
        else if (islower(c))
                return c - 'a' + 10;
        else
                return c - 'A' + 10;
}

void ether_aton(const char *s, struct ether_addr *eap)
{
        unsigned char *ep;
        unsigned int d;

        ep = (u_char *) (eap->ether_addr_octet);

        while (*s) {
                if (*s == ':')
                        s += 1;
                d = xdtoi(*s++);
                if (isxdigit(*s)) {
                        d <<= 4;
                        d |= xdtoi(*s++);
                }
                *ep++ = d;
        }
}

